#include "Shapes.h"

float Shape::intersect( const Vec3& aRayOrigin, const Vec3& aRayDir, bool aPrimaryRay ) const
{
    if (type == Type::Plane)
    {
        const float len = (aPrimaryRay ? eyeDotNormal : (position - aRayOrigin).dot( normal )) / normal.dot( aRayDir );
        return len > 0.003f ? len : -1.0f;
    }
    else if (type == Type::Sphere)
    {
        const Vec3 distance = aRayOrigin - position;
        const float b = distance.dot( aRayDir );
        const float c = distance.dot( distance ) - radius;
        const float d = b * b - c;
        
        return d > 0.003f ? (-b - sqrt( d )) : -1.0f;
    }
    else if (type == Type::Triangle)
    {
        // Algorithm source: http://geomalgorithms.com/a06-_intersect-2.html#intersect_RayTriangle()
        Vec3 vn1 = v1 - v0;
        Vec3 vn2 = v2 - v0;

        float distance = -(aRayOrigin - v0).dot( normal ) / aRayDir.dot( normal );

        if (distance < 0.003f)
        {
            return -1.0f;
        }

        Vec3 hitp = aRayOrigin + aRayDir * distance;

        // Is hit point inside the triangle?
        float uu = vn1.dot( vn1 );
        float uv = vn1.dot( vn2 );
        float vv = vn2.dot( vn2 );
        Vec3 w = hitp - v0;
        float wu = w.dot(vn1);
        float wv = w.dot(vn2);
        float D = uv * uv - uu * vv;
        
        // get and test parametric coords
        float s = (uv * wv - vv * wu) / D;
        
        if (s <= 0.0f || s >= 1.0f)
        {
            return -1.0f;
        }
        
        float t = (uv * wu - uu * wv) / D;
        
        if (t <= 0.0f || (s + t) >= 1.0f)
        {
            return -1.0f;
        }

        return distance;
    }

    return -1.0f;
}

void Shape::getCollisionInfo( const Vec3& aRayOrigin, const Vec3& aRayDir, float aDistance, Vec3& aOutHitpoint, Vec3& aOutNormal ) const
{
    aOutHitpoint = aRayOrigin + aRayDir * aDistance;

    if (type == Type::Sphere)
    {
        aOutNormal = aOutHitpoint - position;
        aOutNormal.normalize();
    }
    else if (type == Type::Plane)
    {
        aOutNormal = normal;
    }
    else if (type == Type::Triangle)
    {
        aOutNormal = normal;
    }
}
